import random

import click

from hyperband_snakemake.search import HbSearch
from hyperband_snakemake.writer import HbWriter


@click.command()
@click.argument('smax', type=int)
@click.argument('eta', type=int)
@click.option('--cost-one-epoch-full-dataset', '-t', type=float,
              help='Cost to train for one epoch on the full dataset')
@click.option('--repetitions', '-r', type=int, default=1,
              help='Number of repetitions for cross-validation')
@click.option('--folds', '-k', type=int, default=5,
              help='Number of folds for cross-validation')
@click.option('--random-seed', type=int, help='Seed for the random generator')
@click.option('--guaranteed-budget', type=int, default=0,
              help='Minimum budget added to each configuration')
@click.option('--output-dir', '-o', type=click.Path(),
              help='Base directory to store the generated configurations')
@click.option('--overwrite', is_flag=True,
              help='Overwrite existing configurations')
@click.option('--template-dir', type=str,
              help='Path to the folder containing the templates')
@click.option('--config-template', default='config',
              help='Path to the configuration template (relative to the template dir.)')
@click.option('--run-template', default='run.sh',
              help='Path to the launch template (relative to the template dir.)')
@click.option('--snakefile-template', default='Snakefile',
              help='Path to the snakefile template (relative to the template dir.)')
def main(output_dir, smax, eta, cost_one_epoch_full_dataset, repetitions, folds,
         random_seed, guaranteed_budget, overwrite, template_dir, config_template,
         run_template, snakefile_template):
    '''
    Uses the hyperband algorithm to generate a structured randomized hyper-parameter search.
    '''
    random.seed(random_seed)

    unit_time = 1.0
    if cost_one_epoch_full_dataset is not None and repetitions is not None and folds is not None:
        epoch_cost = cost_one_epoch_full_dataset * \
            ((1 - 1 / folds) if folds > 1 else 1)
        unit_time = repetitions * folds * epoch_cost

    search = HbSearch(
        smax, eta, unit_time=unit_time, folds=folds,
        repetitions=repetitions, guaranteed_budget=guaranteed_budget
    )
    search.pprint()

    print()
    if output_dir is not None:
        print('Saving random configurations to', output_dir)
        writer = HbWriter(config_template, run_template,
                          snakefile_template, template_dir)
        writer.write_search(search, output_dir, overwrite)
    else:
        print('Not saving configurations (specify target directory with --output-dir)')


if __name__ == '__main__':
    main()




import math


class HbStage:
    def __init__(self, n, r, bracket, search):
        self.n, self.r = n, r
        self.bracket, self.search = bracket, search

    def cost(self):
        return self.search.unit_time * self.n * (self.r + self.search.guaranteed_budget)

    def pprint(self, idx):
        print('    Stage {} - {} configurations each with budget {} (cost: {:.2f})'.format(
            idx, self.n, self.search.guaranteed_budget + self.r, self.cost()
        ))


class HbBracket:
    def __init__(self, eta, s, n, r, search):
        self.s, self.n, self.r, self.eta = s, n, r, eta
        self.search = search
        self.stages = [
            HbStage(
                n=math.floor(n / eta**i),
                r=r * eta**i,
                bracket=self,
                search=search,
            )
            for i in range(0, s + 1)
        ]

    def cost(self):
        return sum(s.cost() for s in self.stages)

    def pprint(self, idx):
        print(f'  Bracket {idx} (cost: {self.cost():.2f})')
        for i, each in enumerate(self.stages):
            each.pprint(i)


class HbSearch:
    def __init__(self, smax, eta, unit_time, folds, repetitions, guaranteed_budget):
        R = eta**smax
        B = R * (smax + 1)

        self.smax, self.eta, self.R, self.B = smax, eta, R, B
        self.unit_time, self.folds, self.repetitions = unit_time, folds, repetitions
        self.guaranteed_budget = guaranteed_budget

        self.brackets = [
            HbBracket(
                eta=eta,
                s=s,
                n=math.ceil(B * eta**s / (R * (s + 1))),
                r=R / eta**s,
                search=self,
            )
            for s in range(smax, -1, -1)
        ]

    def cost(self):
        return sum(b.cost() for b in self.brackets)

    def pprint(self):
        print(f'Hyperband Search (cost: {self.cost():.2f})')
        for i, each in enumerate(self.brackets):
            each.pprint(i)



import os

from jinja2 import Environment, FileSystemLoader, PackageLoader


class HbWriter:
    def __init__(self, config_template, run_template, snakefile_template, template_dir=None):
        self._overwrite_warning = False
        self._config_template = config_template
        self._run_template = run_template
        self._snakefile_template = snakefile_template

        if template_dir is not None:
            loader = FileSystemLoader(template_dir)
        else:
            loader = PackageLoader('hyperband_snakemake', 'templates')

        self._env = Environment(loader=loader, trim_blocks=True, lstrip_blocks=True)

    def render_stage_config(self, index, search, stage):
        tmpl = self._env.get_template(self._config_template)
        rendered = tmpl.render(
            folds=stage.search.folds,
            repetitions=stage.search.repetitions,
            search=search,
            index=index,
        )
        return rendered

    def render_snakefile(self, output_dir, search):
        brackets = [{
            'id': i,
            'max_stage': len(b.stages),
            'num_best': b.stages[-1].n,
            'stages': [{
                'id': j,
                'configs': s.n,
                'budget': int(search.guaranteed_budget + s.r),
                'promote_count': b.stages[j + 1].n if j < len(b.stages) - 1 else 0,
                'promote_epochs': (
                    search.guaranteed_budget + b.stages[j + 1].r
                    if j < len(b.stages) - 1 else None
                ),
            } for j, s in enumerate(b.stages)]
        } for i, b in enumerate(search.brackets)]

        tmpl = self._env.get_template(self._snakefile_template)
        rendered = tmpl.render(
            brackets=brackets, base_dir=output_dir, search=search)
        return rendered

    def render_run(self, search):
        tmpl = self._env.get_template(self._run_template)
        rendered = tmpl.render(search=search)
        return rendered

    @staticmethod
    def _ensure_exists(*parts):
        dd = None
        for p in parts:
            dd = os.path.join(dd, p) if dd is not None else p
            if not os.path.exists(dd):
                os.mkdir(dd)
        return dd

    def _write_to_file(self, string, path, overwrite):
        if isinstance(path, (list, tuple)):
            path = os.path.join(self._ensure_exists(*path[:-1]), path[-1])

        exists = os.path.exists(path)
        if exists and not overwrite:
            if not self._overwrite_warning:
                print('WARNING: not overwriting existing configuration files(s)!')
                print('WARNING: only warning once.')
                self._overwrite_warning = True
        else:
            with open(path, 'w') as f:
                f.write(string)

    def write_search(self, search, output_dir, overwrite=False):
        self._ensure_exists(output_dir)

        self._write_to_file(
            self.render_run(search),
            os.path.join(output_dir, 'run.sh'),
            overwrite
        )

        self._write_to_file(
            self.render_snakefile(output_dir, search),
            os.path.join(output_dir, 'Snakefile'),
            overwrite
        )

        for i, bracket in enumerate(search.brackets):
            for j in range(bracket.stages[0].n):
                self._write_to_file(
                    self.render_stage_config(j, search, bracket.stages[0]),
                    [output_dir, f'bracket-{i}',
                        'stage-0', f'config-{j}', 'config'],
                    overwrite
                )



import setuptools

with open('README.md', 'r') as fh:
    long_description = fh.read()

setuptools.setup(
    name='hyperband-snakemake',
    version='0.0.1',
    author='e-dorigatti',
    author_email='emilio.dorigatti@gmail.com',
    description='Orchestrate hyper-parameters search with snakemake',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/e-dorigatti/hyperband-snakemake/',
    packages=setuptools.find_packages(),
    package_data={'hyperband_snakemake': ['templates/*']},
    install_requires=[
        'click>=7.1.2',
        'jinja2>=2.11.2',
    ],
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6',
)

